/*
 * Copyright 2011 Red Hat, Inc, and individual contributors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.projectodd.stilts.stomplet.container.xa;

import java.util.concurrent.ConcurrentLinkedQueue;

import org.jboss.logging.Logger;
import org.projectodd.stilts.stomp.Acknowledger;
import org.projectodd.stilts.stomp.StompMessage;
import org.projectodd.stilts.stomp.spi.StompSession;
import org.projectodd.stilts.stomplet.Stomplet;

public class PseudoXAStompletTransaction {

    private static Logger log = Logger.getLogger(PseudoXAStompletTransaction.class);

    PseudoXAStompletTransaction(Stomplet stomplet) {
        this.stomplet = stomplet;
    }

    public void setRollbackOnly(boolean rollbackOnly) {
        this.rollbackOnly = rollbackOnly;
    }

    public boolean isRollbackOnly() {
        return this.rollbackOnly;
    }

    boolean isReadOnly() {
        return (this.sentMessages.isEmpty() && this.acks.isEmpty() && this.nacks.isEmpty() );
    }

    public void addSentMessage(StompMessage message, StompSession session) {
        this.sentMessages.add( new MessageSend( message, session)  );
    }

    boolean hasSentMessages() {
        return ! this.sentMessages.isEmpty();
    }

    public void addAck(Acknowledger acknowledger) {
        this.acks.add(  acknowledger );
    }

    boolean hasAcks() {
        return ! this.acks.isEmpty();
    }

    public void addNack(Acknowledger acknowledger) {
        this.nacks.add(acknowledger);
    }

    boolean hasNacks() {
        return ! this.nacks.isEmpty();
    }

    public void commit() {
        for (MessageSend each : sentMessages ) {
            try {
                this.stomplet.onMessage( each.message, each.session );
            } catch (Exception e) {
                log.errorf(e, "Cannot send message: %s", each);
            }
        }

        sentMessages.clear();

        for ( Acknowledger each : this.acks ) {
            try {
                each.ack();
            } catch (Exception e) {
                log.errorf(e, "Cannot ack: %s", each);
            }
        }

        this.acks.clear();

        for ( Acknowledger each : this.nacks ) {
            try {
                each.nack();
            } catch (Exception e) {
                log.errorf(e, "Cannot nack: %s", each);
            }
        }

        this.nacks.clear();
    }

    public void rollback() {
        sentMessages.clear();
        acks.clear();
    }

    private Stomplet stomplet;
    private boolean rollbackOnly;

    private ConcurrentLinkedQueue<MessageSend> sentMessages = new ConcurrentLinkedQueue<MessageSend>();
    private ConcurrentLinkedQueue<Acknowledger> acks = new ConcurrentLinkedQueue<Acknowledger>();
    private ConcurrentLinkedQueue<Acknowledger> nacks = new ConcurrentLinkedQueue<Acknowledger>();


    private static class MessageSend {
        MessageSend(StompMessage message, StompSession session) {
            this.message = message;
            this.session = session;
        }
        StompMessage message;
        StompSession session;
    }
}
